@@ -5,7 +5,7 @@ import math
 import pickle
 from pathlib import Path
 import torch
-from magicsoup.containers import Cell, Protein, Chemistry
+from magicsoup.containers import Cell, Chemistry
 from magicsoup.util import moore_nghbrhd
 from magicsoup.kinetics import Kinetics
 from magicsoup.genetics import Genetics
@@ -16,16 +16,19 @@ class World:
     This is the main object for running the simulation.
     It holds all information and includes all methods for advancing the simulation.
 
-    - `chemistry` The chemistry object that defines molecule species and reactions for this simulation.
-    - `map_size` Size of world map as number of pixels in x- and y-direction
-    - `abs_temp` Absolute temperature in Kelvin will influence the free Gibbs energy calculation of reactions.
-      Higher temperature will give the reaction quotient term higher importance.
-    - `mol_map_init` How to initialize molecule maps (`randn` or `zeros`).
-      `randn` is normally distributed N(10, 1), `zeros` is all zeros.
-    - `start_codons` start codons which start a coding sequence (translation only happens within coding sequences)
-    - `stop_codons` stop codons which stop a coding sequence (translation only happens within coding sequences)
-    - `device` Device to use for tensors (see [pytorch CUDA semantics](https://pytorch.org/docs/stable/notes/cuda.html)).
-      This has to be the same device that is used by `world`.
+    Parameters:
+        chemistry: The chemistry object that defines molecule species and reactions for this simulation.
+        map_size: Size of world map as number of pixels in x- and y-direction.
+        abs_temp: Absolute temperature in Kelvin will influence the free Gibbs energy calculation of reactions.
+            Higher temperature will give the reaction quotient term higher importance.
+        mol_map_init: How to initialize molecule maps (`randn` or `zeros`).
+            `randn` is normally distributed N(10, 1), `zeros` is all zeros.
+        start_codons: start codons which start a coding sequence (translation only happens within coding sequences).
+        stop_codons: stop codons which stop a coding sequence (translation only happens within coding sequences).
+        device: Device to use for tensors (see [pytorch CUDA semantics](https://pytorch.org/docs/stable/notes/cuda.html)).
+            This can be used to move most calculations to a GPU.
+        workers: Number of multiprocessing workers to use.
+            These are used to parallelize some calculations that can only be done on the CPU.
 
     Most attributes on this class describe the current state of molecules and cells.
     Whenever molecules are listed or represented in one dimension, they are ordered the same way as in `chemistry.molecules`.
@@ -35,69 +38,89 @@ class World:
     But whenever an operation modifies the number of cells (like `world.kill_cells()` or `world.replicate_cells()`),
     cells get new indexes. Here are the most important attributes:
 
-    - `cells` A list of cell objects. These cell objects hold e.g. each cell's genome and proteome.
-    - `cell_map` Boolean 2D tensor referencing which pixels are occupied by a cell.
-      Dimension 0 represents the x, dimension 1 y.
-    - `molecule_map` Float 3D tensor describing how many molecules (in mol) of each molecule species exist on every pixel in this world.
-      Dimension 0 describes the molecule species. They are in the same order as `chemistry.molecules`.
-      Dimension 1 represents x, dimension 2 y.
-      So, `world.molecule_map[0, 1, 2]` is number of molecules of the 0th molecule species on pixel 1, 2.
-    - `cell_molecules` Float 2D tensor describing the number of molecules (in mol) for each molecule species in each cell.
-      Dimension 0 is the cell index. It is the same as in `world.cells` and the same as on a cell object (`cell.idx`).
-      Dimension 1 describes the molecule species. They are in the same order as `chemistry.molecules`.
-      So, `world.cell_molecules[0, 1]` represents how many mol of the 1st molecule species the 0th cell contains.
-    - `cell_survival` Integer 1D tensor describing how many time steps each cell survived.
-      Cells are in the same as in `world.cells` and the same as on a cell object (`cell.idx`).
-    - `cell_divisions` Integer 1D tensor describing how many times each cell replicated.
-      Cells are in the same as in `world.cells` and the same as on a cell object (`cell.idx`).
+    Attributes:
+        cells: A list of cell objects. These cell objects hold e.g. each cell's genome and proteome.
+        cell_map: Boolean 2D tensor referencing which pixels are occupied by a cell.
+            Dimension 0 represents the x, dimension 1 y.
+        molecule_map: Float 3D tensor describing how many molecules (in mol) of each molecule species exist on every pixel in this world.
+            Dimension 0 describes the molecule species. They are in the same order as `chemistry.molecules`.
+            Dimension 1 represents x, dimension 2 y.
+            So, `world.molecule_map[0, 1, 2]` is number of molecules of the 0th molecule species on pixel 1, 2.
+        cell_molecules: Float 2D tensor describing the number of molecules (in mol) for each molecule species in each cell.
+            Dimension 0 is the cell index. It is the same as in `world.cells` and the same as on a cell object (`cell.idx`).
+            Dimension 1 describes the molecule species. They are in the same order as `chemistry.molecules`.
+            So, `world.cell_molecules[0, 1]` represents how many mol of the 1st molecule species the 0th cell contains.
+        cell_survival: Integer 1D tensor describing how many time steps each cell survived.
+            This tensor is for monitoring and doesn't have any other effect.
+            Cells are in the same as in `world.cells` and the same as on a cell object (`cell.idx`).
+        cell_divisions: Integer 1D tensor describing how many times each cell replicated.
+            This tensor is for monitoring and doesn't have any other effect.
+            Cells are in the same as in `world.cells` and the same as on a cell object (`cell.idx`).
 
-    Then, there are methods for advancing the simulation:
+    Methods for advancing the simulation and to use during a simulation:
 
-    - `add_random_cells` add new cells and place them randomly on the map
-    - `replicate_cells` replicate existing cells
-    - `update_cells` update existing cells if their genome has changed
-    - `kill_cells` kill existing cells
-    - `move_cells` move existing cells to a random position in their Moore's neighborhood
-    - `diffuse_molecules` let molecules diffuse and permeate by one time step
-    - `degrade_molecules` let molecules degrade by one time step
-    - `increment_cell_survival` increment `world.cell_survival` by 1
-    - `enzymatic_activity` let cell proteins work for one time step
-      
-    If you want to get a cell with all information about its contents and its current environment use `world.get_cell()`.
+    - [add_random_cells()][magicsoup.world.World.add_random_cells] add new cells and place them randomly on the map
+    - [replicate_cells()][magicsoup.world.World.replicate_cells] replicate existing cells
+    - [update_cells()][magicsoup.world.World.update_cells] update existing cells if their genome has changed
+    - [kill_cells()][magicsoup.world.World.kill_cells] kill existing cells
+    - [move_cells()][magicsoup.world.World.move_cells] move existing cells to a random position in their Moore's neighborhood
+    - [diffuse_molecules()][magicsoup.world.World.diffuse_molecules] let molecules diffuse and permeate by one time step
+    - [degrade_molecules()][magicsoup.world.World.degrade_molecules] let molecules degrade by one time step
+    - [increment_cell_survival()][magicsoup.world.World.increment_cell_survival] increment `world.cell_survival` by 1
+    - [enzymatic_activity()][magicsoup.world.World.enzymatic_activity] let cell proteins work for one time step
+
+    If you want to get a cell with all information about its contents and its current environment use [get_cell()][magicsoup.world.World.get_cell].
     During the simulation you should however work directly with the tensors mentioned above for performance reasons.
 
     Furthermore, there are methods for saving and loading a simulation.
-    For any new simulation use `world.save()` once to save the whole world object (with chemistry, genetics, kinetics)
-    to a pickle file. You can restore it with `World.from_file()` later on.
-    Then, to save the world's state you can use `world.save_state()`.
+    For any new simulation use [save()][magicsoup.world.World.save] once to save the whole world object (with chemistry, genetics, kinetics)
+    to a pickle file. You can restore it with [from_file()][magicsoup.world.World.from_file] later on.
+    Then, to save the world's state you can use [save_state()][magicsoup.world.World.save_state].
     This is a quick, lightweight save, but it only saves things that change during the simulation.
-    Use `world.load_state()` to re-load a certain state.
+    Use [load_state()][magicsoup.world.World.load_state] to re-load a certain state.
 
-    Finally, the `world` object carries `world.genetics`, `world.kinetics`, and `world.genetics.chemistry`
+    Finally, the `world` object carries `world.genetics`, `world.kinetics`, and `world.chemistry`
     (which is just a reference to the chemistry object that was used when initializing `world`).
     Usually, you don't need to touch them.
     But, if you want to override them or look into some details, see the docstrings of their classes for more information.
     """
 
+    # TODO: Cell positions could be maintained in a tensor only.
+    #       That way I could keep them on the GPU
+    #       Often I move them from CPU to GPU, just to add them to cell.position
+    #       But their actual use is always on the GPU: indexing on self.cell_map
+
     def __init__(
         self,
         chemistry: Chemistry,
-        map_size=128,
-        abs_temp=310.0,
-        mol_map_init="randn",
+        map_size: int = 128,
+        abs_temp: float = 310.0,
+        mol_map_init: str = "randn",
         start_codons: tuple[str, ...] = ("TTG", "GTG", "ATG"),
         stop_codons: tuple[str, ...] = ("TGA", "TAG", "TAA"),
-        device="cpu",
+        device: str = "cpu",
+        workers: int = 2,
     ):
         if not torch.cuda.is_available():
             device = "cpu"
 
         self.device = device
+        self.workers = workers
         self.map_size = map_size
         self.abs_temp = abs_temp
+        self.chemistry = chemistry
 
         self.genetics = Genetics(
-            chemistry=chemistry, start_codons=start_codons, stop_codons=stop_codons,
+            start_codons=start_codons,
+            stop_codons=stop_codons,
+        )
+
+        self.kinetics = Kinetics(
+            molecules=chemistry.molecules,
+            reactions=chemistry.reactions,
+            abs_temp=abs_temp,
+            device=self.device,
+            workers=self.workers,
         )
 
         mol_degrads: list[float] = []
@@ -116,26 +139,19 @@ class World:
         self._permeation = permeation
 
         self._nghbrhd_map = {
-            (x, y): torch.tensor(moore_nghbrhd(x, y, map_size))
+            (x, y): torch.tensor(moore_nghbrhd(x, y, map_size)).to(self.device)
             for x, y in product(range(map_size), range(map_size))
         }
 
         self.cells: list[Cell] = []
-        self.kinetics = Kinetics(
-            molecules=chemistry.molecules, abs_temp=abs_temp, device=self.device,
+        self.cell_map: torch.Tensor = torch.zeros(map_size, map_size).to(device).bool()
+        self.cell_survival: torch.Tensor = torch.zeros(0).to(device).int()
+        self.cell_divisions: torch.Tensor = torch.zeros(0).to(device).int()
+        self.cell_molecules: torch.Tensor = torch.zeros(0, self.n_molecules).to(device)
+        self.molecule_map: torch.Tensor = self._get_molecule_map(
+            n=self.n_molecules, size=map_size, init=mol_map_init
         )
 
-        self.cell_map = torch.zeros(map_size, map_size).to(device).bool()
-        self.cell_survival = torch.zeros(0).to(device).int()
-        self.cell_divisions = torch.zeros(0).to(device).int()
-        self.cell_molecules = torch.zeros(0, self.n_molecules).to(device)
-        self.molecule_map = self._get_molecule_map(
-            n=self.n_molecules, size=self.map_size, init=mol_map_init
-        ).to(self.device)
-
-    # TODO: does it make sense to keep cell indexes constant (even after deleting cells)
-    #       and only adding new indexes?
-
     def get_cell(
         self,
         by_idx: Optional[int] = None,
@@ -143,17 +159,23 @@ class World:
         by_position: Optional[tuple[int, int]] = None,
     ) -> Cell:
         """
-        Get a cell with information about its current environment
-        
-        - `by_idx` get cell by cell index (`cell.idx`)
-        - `by_label` get cell by cell label (`cell.label`)
-        - `by_position` get cell by position (x, y)
+        Get a cell with information about its current environment.
+        Raises `ValueError` if cell was not found.
+
+        Parameters:
+            by_idx: get cell by cell index (`cell.idx`)
+            by_label: get cell by cell label (`cell.label`)
+            by_position: get cell by position (x, y)
+
+        Returns:
+            The searched cell.
 
         When accessing `world.cells` directly, the cell object will not have all information.
         For performance reasons most cell attributes are maintained in tensors during the simulation.
-        Only genome and proteome are maintained during the simulation.
+        Only index, genome and position are kept up-to-date in the cell object during the simulation.
         When you call `world.get_cell()` all missing information will be added to the object.
         """
+        idx = -1
         if by_idx is not None:
             idx = by_idx
         if by_label is not None:
@@ -172,88 +194,88 @@ class World:
         cell = self.cells[idx]
         cell.int_molecules = self.cell_molecules[idx, :]
         cell.ext_molecules = self.molecule_map[:, cell.position[0], cell.position[1]]
-        cell.n_survived_steps = self.cell_survival[idx].item()
-        cell.n_replications = self.cell_divisions[idx].item()
+        cell.n_survived_steps = int(self.cell_survival[idx].item())
+        cell.n_replications = int(self.cell_divisions[idx].item())
+
+        (cdss,) = self.genetics.translate_genomes(genomes=[cell.genome])
+        cell.proteome = self.kinetics.get_proteome(proteome=cdss)
+
         return cell
 
     def add_random_cells(self, genomes: list[str]) -> list[int]:
         """
         Create new cells and place them on randomly on the map.
+        All lists and tensors that reference cells will be updated.
 
-        - `genomes` list of genomes of the newly added cells
+        Parameters:
+            genomes: List of genomes of the newly added cells
 
-        Returns the indexes of successfully added cells.
+        Returns:
+            The indexes of successfully added cells.
 
         Each cell will be placed randomly on the map and receive half the molecules of the pixel where it was added.
         If there are less pixels left on the cell map than cells you want to add,
         only the remaining pixels will be filled with new cells.
-
-        All lists and tensors that reference cells will be updated.
         """
         genomes = [d for d in genomes if len(d) > 0]
-        if len(genomes) == 0:
-            return []
-
-        xs, ys = self._find_free_random_positions(n_cells=len(genomes))
-        if len(xs) == 0:
-            return []
-
-        prot_lens = []
-        new_idxs = []
-        new_params: list[tuple[int, int, Protein]] = []
-        next_idx = len(self.cells)
-        n_new_cells = 0
-        new_xs = []
-        new_ys = []
-        for genome in genomes:
-            proteome = self.genetics.get_proteome(seq=genome)
-            n_proteins = len(proteome)
-            if n_proteins == 0:
-                continue
-
-            prot_lens.append(n_proteins)
-            new_idxs.append(next_idx)
-            for prot_i, prot in enumerate(proteome):
-                new_params.append((next_idx, prot_i, prot))
-
-            x = xs[n_new_cells]
-            y = ys[n_new_cells]
-            new_xs.append(x)
-            new_ys.append(y)
-
-            cell = Cell(idx=next_idx, genome=genome, proteome=proteome, position=(x, y))
-            self.cells.append(cell)
-
-            next_idx += 1
-            n_new_cells += 1
-
+        n_new_cells = len(genomes)
         if n_new_cells == 0:
             return []
 
+        xs, ys = self._find_free_random_positions(n_cells=n_new_cells)
+        n_avail_pos = len(xs)
+        if n_avail_pos == 0:
+            return []
+
+        if n_avail_pos < n_new_cells:
+            n_new_cells = n_avail_pos
+            random.shuffle(genomes)
+            genomes = genomes[:n_new_cells]
+
+        proteomes = self.genetics.translate_genomes(genomes=genomes)
+        proteomes = [d for d in proteomes if len(d) > 0]
+        n_new_cells = len(proteomes)
+        if n_new_cells == 0:
+            return []
+
+        xs = xs[:n_new_cells]
+        ys = ys[:n_new_cells]
+
+        n_cells = len(self.cells)
+        new_idxs = list(range(n_cells, n_cells + n_new_cells))
+        for cell_i, genome, x, y in zip(new_idxs, genomes, xs, ys):
+            cell = Cell(idx=cell_i, genome=genome, proteome=[], position=(x, y))
+            self.cells.append(cell)
+
         self.cell_survival = self._expand(t=self.cell_survival, n=n_new_cells, d=0)
         self.cell_divisions = self._expand(t=self.cell_divisions, n=n_new_cells, d=0)
         self.cell_molecules = self._expand(t=self.cell_molecules, n=n_new_cells, d=0)
+
+        n_max_prots = max(len(d) for d in proteomes)
+        self.kinetics.increase_max_proteins(max_n=n_max_prots)
         self.kinetics.increase_max_cells(by_n=n_new_cells)
-        self.kinetics.increase_max_proteins(max_n=max(prot_lens))
-        self.kinetics.set_cell_params(cell_prots=new_params)
+        self.kinetics.set_cell_params(cell_idxs=new_idxs, proteomes=proteomes)
 
         # occupy positions
-        self.cell_map[new_xs, new_ys] = True
+        self.cell_map[xs, ys] = True
 
         # cell is picking up half the molecules of the pxl it is born on
-        pickup = self.molecule_map[:, new_xs, new_ys] * 0.5
+        pickup = self.molecule_map[:, xs, ys] * 0.5
         self.cell_molecules[new_idxs, :] += pickup.T
-        self.molecule_map[:, new_xs, new_ys] -= pickup
+        self.molecule_map[:, xs, ys] -= pickup
 
         return new_idxs
 
     def replicate_cells(self, parent_idxs: list[int]) -> list[tuple[int, int]]:
         """
-        Bring cells to replicate
+        Bring cells to replicate.
+        All lists and tensors that reference cells will be updated.
 
-        - `parent_idxs` Cell indexes of the cells that should replicate.
+        Parameters:
+            parent_idxs: Cell indexes of the cells that should replicate.
 
-        Returns a list of tuples of parent and child cell indexes for all successfully replicated cells.
+        Returns:
+            A list of tuples of parent and child cell indexes for all successfully replicated cells.
 
         In this simulation the cell that was brought to replicate is the parent.
         It still lives on the same pixel.
@@ -264,8 +286,6 @@ class World:
 
         Each child will be placed randomly next to its parent (Moore's neighborhood).
         If every pixel in the parent's Moore's neighborhood is taken the cell will not replicate.
-
-        All lists and tensors that reference cells will be updated.
         """
         if len(parent_idxs) == 0:
             return []
@@ -295,62 +315,52 @@ class World:
         """
         Update existing cells with new genomes.
 
-        - `genome_idx_pairs` list of tuples of genomes and cell indexes
+        Parameters:
+            genome_idx_pairs: List of tuples of genomes and cell indexes
 
         The indexes refer to the index of each cell that is changed.
         The genomes refer to the genome of each cell that is changed.
         `world.cells` will be updated with new genomes and proteomes.
         """
-        # TODO: 3 steps where only 1 proc is 100% active total about 25s
-        #       then 1 step where all procs are 50% active (probably enzymatic_activity())
-        #       total about 7k cells with avg genome size 4000
-        #       needs to speed up....
-
         if len(genome_idx_pairs) == 0:
             return
 
-        kill_idxs: list[int] = []
-        prot_lens: list[int] = []
-        set_params: list[tuple[int, int, Protein]] = []
-        unset_params: list[tuple[int, int]] = []
+        kill_idxs = []
+        transl_idxs = []
+        genomes = []
         for genome, idx in genome_idx_pairs:
-            if len(genome) == 0:
+            if len(genome) > 0:
+                genomes.append(genome)
+                transl_idxs.append(idx)
+            else:
                 kill_idxs.append(idx)
-                continue
 
-            cell = self.cells[idx]
-            newprot = self.genetics.get_proteome(seq=genome)
-            if len(newprot) == 0:
+        proteomes = self.genetics.translate_genomes(genomes=genomes)
+
+        set_idxs = []
+        set_proteomes = []
+        for proteome, idx in zip(proteomes, transl_idxs):
+            if len(proteome) > 0:
+                set_proteomes.append(proteome)
+                set_idxs.append(idx)
+            else:
                 kill_idxs.append(idx)
-                continue
 
-            oldprot = cell.proteome
-            n_new = len(newprot)
-            n_old = len(oldprot)
-            n = min(n_old, n_new)
-            for pi, (np, op) in enumerate(zip(newprot[:n], oldprot[:n])):
-                if np != op:
-                    set_params.append((idx, pi, np))
-            if n_old > n_new:
-                unset_params.extend((idx, i) for i in range(n, n_old))
-
-            cell.proteome = newprot
-            cell.genome = genome
-            prot_lens.append(n_new)
-
-        self.kinetics.increase_max_proteins(max_n=max(prot_lens))
-        self.kinetics.set_cell_params(cell_prots=set_params)
-        self.kinetics.unset_cell_params(cell_prots=unset_params)
+        max_prots = max(len(d) for d in set_proteomes)
+        self.kinetics.increase_max_proteins(max_n=max_prots)
+        self.kinetics.set_cell_params(cell_idxs=set_idxs, proteomes=set_proteomes)
         self.kill_cells(cell_idxs=kill_idxs)
 
     def kill_cells(self, cell_idxs: list[int]):
         """
-        Remove existing cells
-        
-        - `cell_idxs` indexes of the cells that should die
-        
+        Remove existing cells.
+        All lists and tensors referencing cells will be updated.
+
+        Parameters:
+            cell_idxs: Indexes of the cells that should die
+
         Cells that are killed dump their molecule contents onto the pixel they used to live on.
-        
+
         Cells will be removed from all lists and tensors that reference cells.
         Thus, after killing cells the index of some living cells will be updated.
         E.g. if there are 10 cells and you kill the cell with index 8 (the 9th cell),
@@ -365,7 +375,7 @@ class World:
         spillout = self.cell_molecules[cell_idxs, :]
         self.molecule_map[:, xs, ys] += spillout.T
 
-        keep = torch.ones(self.cell_survival.size(0), dtype=torch.bool)
+        keep = torch.ones(self.cell_survival.size(0), dtype=torch.bool).to(self.device)
         keep[cell_idxs] = False
         self.cell_survival = self.cell_survival[keep]
         self.cell_divisions = self.cell_divisions[keep]
@@ -383,12 +393,12 @@ class World:
 
     def move_cells(self, cell_idxs: list[int]):
         """
-        Move cells to a random position in their Moore's neighborhood
-
-        - `cell_idxs` indexes of cells that should be moved
-        
+        Move cells to a random position in their Moore's neighborhood.
         If every pixel in the cells' Moore neighborhood is taken the cell will not be moved.
         `world.cell_map` will be updated.
+
+        Parameters:
+            cell_idxs: Indexes of cells that should be moved
         """
         if len(cell_idxs) == 0:
             return
@@ -398,10 +408,9 @@ class World:
 
     def enzymatic_activity(self):
         """
-        Catalyze reactions for one time step
-
+        Catalyze reactions for one time step.
         This includes molecule transport into or out of the cell.
-        `world.molecule_map` and `cell_molecules` are updated.
+        `world.molecule_map` and `world.cell_molecules` are updated.
         """
         if len(self.cells) == 0:
             return
@@ -418,17 +427,24 @@ class World:
         """
         Let molecules in molecule map diffuse and permeate through cell membranes
         by one time step.
+        `world.molecule_map` and `world.cell_molecules` are updated.
 
         By how much each molcule species diffuses around the world map and permeates
         into or out of cells if defined on the `Molecule` objects itself.
         See `Molecule` for more information.
-
-        `world.molecule_map` and `cell_molecules` are updated.
         """
+        n_pxls = self.map_size**2
         for mol_i, diffuse in enumerate(self._diffusion):
+            total_before = self.molecule_map[mol_i].sum()
             before = self.molecule_map[mol_i].unsqueeze(0).unsqueeze(1)
             after = diffuse(before)
             self.molecule_map[mol_i] = torch.squeeze(after, 0).squeeze(0)
+            total_after = self.molecule_map[mol_i].sum()
+
+            # attempt to fix the problem that convolusion makes a small amount of
+            # molecules appear or disappear (I think because floating point)
+            self.molecule_map[mol_i] += (total_before - total_after) / n_pxls
+            self.molecule_map[mol_i] = self.molecule_map[mol_i].clamp(0.0)
 
         if len(self.cells) == 0:
             return
@@ -448,11 +464,10 @@ class World:
     def degrade_molecules(self):
         """
         Degrade molecules in world map and cells by one time step.
-        
+        `world.molecule_map` and `world.cell_molecules` are updated.
+
         How quickly each molecule species degrades depends on its half life
         which is defined on the `Molecule` object of that species.
-        
-        `world.molecule_map` and `cell_molecules` are updated.
         """
         for mol_i, degrad in enumerate(self._mol_degrads):
             self.molecule_map[mol_i] *= degrad
@@ -461,21 +476,23 @@ class World:
     def increment_cell_survival(self):
         """
         Increment `world.cell_survival` by 1.
+        This is for monitoring and doesn't have any other effect.
         """
         idxs = list(range(len(self.cells)))
         self.cell_survival[idxs] += 1
 
-    def save(self, rundir: Path, name="world.pkl"):
+    def save(self, rundir: Path, name: str = "world.pkl"):
         """
         Write whole world object to pickle file
 
-        - `rundir` directory of the pickle file
-        - `name` name of the pickle file
-        
+        Parameters:
+            rundir: Directory of the pickle file
+            name: Name of the pickle file
+
         This is a big and slow save.
         It saves everything needed to restore the world with its chemistry and genetics.
-        Use `World.from_file()` to restore it.
-        For a small and quick save use `world.save_state()`.
+        Use [from_file()][magicsoup.world.World.from_file] to restore it.
+        For a small and quick save use [save_state()][magicsoup.world.World.save_state].
         """
         # TODO: make this JSON, txt, (except for tensors), to make it possible
         #       to continue using a different language
@@ -485,14 +502,17 @@ class World:
             pickle.dump(self, fh)
 
     @classmethod
-    def from_file(self, rundir: Path, name="world.pkl") -> "World":
+    def from_file(self, rundir: Path, name: str = "world.pkl") -> "World":
         """
         Restore previously saved world from pickle file.
+        The file had to be saved with [save()][magicsoup.world.World.save].
 
-        - `rundir` directory of the pickle file
-        - `name` name of the pickle file
+        Parameters:
+            rundir: Directory of the pickle file
+            name: Name of the pickle file
 
-        The file had to be saved with `world.save()`.
+        Returns:
+            A new `world` instance.
         """
         with open(rundir / name, "rb") as fh:
             return pickle.load(fh)
@@ -501,13 +521,14 @@ class World:
         """
         Save current state only
 
-        - `statedir` directory to store files in (there are multiple files per state)
-        
+        Parameters:
+            statedir: Directory to store files in (there are multiple files per state)
+
         This is a small and quick save.
         It only saves things which change during the simulation.
-        Restore a certain state with `world.load_state()`.
-        
-        To restore a world object you need to save it at least once with `world.save()`.
+        Restore a certain state with [load_state()][magicsoup.world.World.load_state].
+
+        To restore a whole `world` object you need to save it at least once with [save()][magicsoup.world.World.save].
         Then, `world.save_state()` can be used to save different states of that world object.
         """
         statedir.mkdir(parents=True, exist_ok=True)
@@ -517,23 +538,23 @@ class World:
         torch.save(self.cell_survival, statedir / "cell_survival.pt")
         torch.save(self.cell_divisions, statedir / "cell_divisions.pt")
 
-        with open(statedir / "cells.fasta", "w") as fh:
+        with open(statedir / "cells.fasta", "w", encoding="utf-8") as fh:
             lines = [
                 f">{d.idx} {d.label} ({d.position[0]},{d.position[1]})\n{d.genome}"
                 for d in self.cells
             ]
             fh.write("\n".join(lines))
 
-    def load_state(self, statedir: Path, update_cell_params=True):
+    def load_state(self, statedir: Path, update_cell_params: bool = True):
         """
         Load a saved world state.
+        The state had to be saved with [save_state()][magicsoup.world.World.save_state] previously.
 
-        - `statedir` directory that contains all files of that state
-        - `update_cell_params` whether to update cell parameters as well.
-          If you are only interested in the cells' genomes and molecules
-          you can set this to `False` to make loading states faster.
-
-        The state had to be saved with `world.save_state()` previously.
+        Parameters:
+            statedir: Directory that contains all files of that state
+            update_cell_params: Whether to update cell parameters as well.
+                If you are only interested in the cells' genomes and molecules
+                you can set this to `False` to make loading states faster.
         """
         cell_molecules: torch.Tensor = torch.load(statedir / "cell_molecules.pt")
         self.cell_molecules = cell_molecules.to(self.device)
@@ -546,7 +567,7 @@ class World:
         cell_divisions: torch.Tensor = torch.load(statedir / "cell_divisions.pt")
         self.cell_divisions = cell_divisions.to(self.device).int()
 
-        with open(statedir / "cells.fasta", "r") as fh:
+        with open(statedir / "cells.fasta", "r", encoding="utf-8") as fh:
             text: str = fh.read()
             entries = [d.strip() for d in text.split(">") if len(d.strip()) > 0]
 
@@ -572,7 +593,7 @@ class World:
             x, y = cell.position
             nghbrhd = self._nghbrhd_map[(x, y)]
             pxls = nghbrhd[~self.cell_map[nghbrhd[:, 0], nghbrhd[:, 1]]]
-            n = len(pxls)
+            n = pxls.size(0)
 
             if n == 0:
                 continue
@@ -595,7 +616,7 @@ class World:
             x, y = parent.position
             nghbrhd = self._nghbrhd_map[(x, y)]
             pxls = nghbrhd[~self.cell_map[nghbrhd[:, 0], nghbrhd[:, 1]]]
-            n = len(pxls)
+            n = pxls.size(0)
 
             if n == 0:
                 continue
@@ -614,11 +635,12 @@ class World:
     def _find_free_random_positions(self, n_cells: int) -> tuple[list[int], list[int]]:
         # available spots on map
         pxls = torch.argwhere(~self.cell_map)
-        if n_cells > len(pxls):
-            n_cells = len(pxls)
+        n_pxls = pxls.size(0)
+        if n_cells > n_pxls:
+            n_cells = n_pxls
 
         # place cells on map
-        idxs = random.sample(range(len(pxls)), k=n_cells)
+        idxs = random.sample(range(n_pxls), k=n_cells)
         chosen = pxls[idxs]
         xs, ys = chosen.T.tolist()
         return xs, ys
@@ -626,9 +648,9 @@ class World:
     def _get_molecule_map(self, n: int, size: int, init: str) -> torch.Tensor:
         args = [n, size, size]
         if init == "zeros":
-            return torch.zeros(*args)
+            return torch.zeros(*args).to(self.device)
         if init == "randn":
-            return torch.abs(torch.randn(*args) + 10.0)
+            return torch.abs(torch.randn(*args) + 10.0).to(self.device)
         raise ValueError(
             f"Didnt recognize mol_map_init={init}."
             " Should be one of: 'zeros', 'randn'."
@@ -670,7 +692,7 @@ class World:
             [a, a, a],
             [a, b, a],
             [a, a, a],
-        ]]])
+        ]]]).to(self.device)
         # fmt: on
 
         conv = torch.nn.Conv2d(
@@ -692,6 +714,11 @@ class World:
         return torch.cat([t, zeros], dim=d)
 
     def __repr__(self) -> str:
-        clsname = type(self).__name__
-        return "%s(map_size=%r,abs_temp=%r)" % (clsname, self.map_size, self.abs_temp,)
-
+        kwargs = {
+            "map_size": self.map_size,
+            "abs_temp": self.abs_temp,
+            "device": self.device,
+            "workers": self.workers,
+        }
+        args = [f"{k}:{repr(d)}" for k, d in kwargs.items()]
+        return f"{type(self).__name__}({','.join(args)})"
